#!/usr/bin/perl

use strict;

use Sendmail::Milter;
use Getopt::Long;
use File::Basename;
use Sys::Syslog qw(:DEFAULT setlogsock);
use Sendmail::AccessDB;
use DB_File;
use Fcntl ':flock';
use Time::Local;
use Socket;

BEGIN {
    push(@INC, '/usr/local/milters/verify-addr-smf.pl', '/usr/local/lib/perl');
}

use check_mail_address;

my($version) = '0.05';
my($author) = 'Victor Ustugov aka corvax';
my($email) = 'victor@corvax.kiev.ua';
my($info) = "verify mail address perl mail filter $version 2003 by $author $email";

my(%defaults) = (
    'sendmail_config'		=> '/etc/mail/sendmail.cf:/etc/sendmail.cf',
    'syslog_facility'		=> 'local4',
    'sender_tempfail_accept'	=> 0,
    'rcpt_tempfail_accept'	=> 0,
    'max_attempts'		=> 1,
    'use_vrfy'			=> 0,
    'access_db'			=> '/etc/mail/access.db',
    'cache_file'		=> '/var/milters/verify-addr-smf.db',
    'skip_hosts'		=> 0,
    'delay_checks'		=> 0
);

my(%defaults_ttl) = (
    'sender_ok'			=> '3600',
    'sender_fail'		=> '3600',
    'sender_defer'		=> '3600',
    'recipient_ok'		=> '3600',
    'recipient_fail'		=> '3600',
    'recipient_defer'		=> '3600'
);

my($script_name) = basename($0);

my(%syslog_param) = (
    'socket' => 'unix',
    'ident' => $script_name,
    'options' => 'cons,pid',
#    'facility' => 'local4',
    'facility' => 'mail',
    'priority' => 'notice',
#  logformat  :
# %mailbox	-  mailbox
# %domain	-  
# %SMTP_host	- primary MX
# %SMTP_command	- ,  MTA  
# %SMTP_answer	-  MTA
# %check_result	-   
#    'logformat' => 'check %mailbox@%domain: %SMTP_command: %SMTP_answer; %check_result'
#    'logformat' => '%check_result (%mailbox@%domain: %SMTP_command => %SMTP_answer)'
    'logformat' => '%check_result (%mailbox@%domain: %SMTP_host => %SMTP_command => %SMTP_answer)'
);

my($verbose, $max_attempts, $sender_tempfail_accept, $rcpt_tempfail_accept, $access_db,
    $skip_authen, $skip_relay_from, $skip_hosts,
    $delay_checks, $clear_cache, $delete_cache, $delete_cache_expired, $dump_cache, $cache_file, %cache_ttl);
my($use_vrfy) = '';

sub lookup_access_db() {
    my($address, $qualifier, $type) = @_;
    $address =~ s/^\s*\<//;
    $address =~ s/\>\s*$//;

    foreach my $file (split(/:/, $access_db)) {
	print(STDERR "lookup_access_db: search ".($qualifier ? "$qualifier:" : '')."$address (type '$type') in '$file'\n") if ($verbose);
	my($key, $value);
	$key = $address;
	if ($type eq 'mail') {
	    unless ($value = Sendmail::AccessDB::lookup($key, 'type'=>'', 'qualifier'=>$qualifier, 'file'=>$file)) {
		my($mailbox, $domain) = ($address =~ /\s*(.+)\@(.+)\s*$/);
	        $key = '@'.$domain;
		unless ($value = Sendmail::AccessDB::lookup($key, 'type'=>'', 'qualifier'=>$qualifier, 'file'=>$file)) {
		    $key = $address;
		    unless ($value = Sendmail::AccessDB::lookup($key, 'type'=>'mail', 'qualifier'=>$qualifier, 'file'=>$file)) {
		        $key = '';
			$value = Sendmail::AccessDB::lookup($key, 'type'=>'mail', 'qualifier'=>$qualifier, 'file'=>$file);
		    }
		}
	    }
	} else {
	    if ($qualifier) {
		($value = Sendmail::AccessDB::lookup($key, 'type'=>$type, 'qualifier'=>$qualifier, 'file'=>$file));
	    } else {
		($value = Sendmail::AccessDB::lookup($key, 'type'=>$type, 'file'=>$file));
	    }
	}
	if ($verbose) {
	    if ($value) {
		print(STDERR "lookup_access_db: found '$value'\n");
	    } else {
		print(STDERR "lookup_access_db: key not found\n");
	    }
	}
	return(($key, $value)) if ($value);
    }
    return('');
}

sub save_cache() {
    my($address, $value) = @_;

    my(%database);
    my($db) = tie(%database, 'DB_File', $cache_file, O_CREAT | O_RDWR, 0644);
    unless ($db) {
	my($msg) = "couldn't init DB '$cache_file': $!";
	print(STDERR "$msg\n") if ($verbose);
	syslog($syslog_param{'priority'}, $msg);
    } else {
	my($fd) = $db->fd();
	unless (open(DATAFILE, "+<&=$fd")) {
	    my($msg) = "couldn't safety open DB '$cache_file': $!";
	    print(STDERR "$msg\n") if ($verbose);
	    syslog($syslog_param{'priority'}, $msg);
	} else {
	    unless (flock(DATAFILE, LOCK_EX)) {
		my($msg) = "couldn't lock DB '$cache_file': $!";
	        print(STDERR "$msg\n") if ($verbose);
		syslog($syslog_param{'priority'}, $msg);
	    } else {
		my($sec, $min, $hour, $mday, $month, $year) = (localtime)[0,1,2,3,4,5];
		$year += 1900;
		$month++;
		$database{$address} = sprintf("%04d%02d%02d%02d%02d%02d\t%s",
		    $year, $month, $mday, $hour, $min, $sec, $value);
	    }
	    $db->sync();
	    close(DATAFILE);
	}
	undef($db);
	untie(%database);
    }
}

sub clear_cache() {
    print(STDERR "clear cache file '$cache_file'\n") if ($verbose);
    my(%database, $db);
    my($db) = tie(%database, 'DB_File', $cache_file, O_CREAT | O_RDWR, 0644);
    unless ($db) {
	my($msg) = "couldn't init DB '$cache_file': $!";
	print(STDERR "$msg\n") if ($verbose);
	syslog($syslog_param{'priority'}, $msg);
    } else {
	my($fd) = $db->fd();
	unless (open(DATAFILE, "+<&=$fd")) {
	    my($msg) = "couldn't safety open DB '$cache_file': $!";
	    print(STDERR "$msg\n") if ($verbose);
	    syslog($syslog_param{'priority'}, $msg);
	} else {
	    unless (flock(DATAFILE, LOCK_EX)) {
		my($msg) = "couldn't lock DB '$cache_file': $!";
	        print(STDERR "$msg\n") if ($verbose);
		syslog($syslog_param{'priority'}, $msg);
	    } else {
		%database = ();
	    }
	    $db->sync();
	    close(DATAFILE);
	}
	undef($db);
	untie(%database);
    }
}

sub delete_cache() {
    my($addr) = @_;
    print(STDERR "delete address '$addr' from cache file '$cache_file'\n") if ($verbose);
    my(%database);
    my($db) = tie(%database, 'DB_File', $cache_file, O_CREAT | O_RDWR, 0644);
    unless ($db) {
	my($msg) = "couldn't init DB '$cache_file': $!";
	print(STDERR "$msg\n") if ($verbose);
	syslog($syslog_param{'priority'}, $msg);
    } else {
	my($fd) = $db->fd();
	unless (open(DATAFILE, "+<&=$fd")) {
	    my($msg) = "couldn't safety open DB '$cache_file': $!";
	    print(STDERR "$msg\n") if ($verbose);
	    syslog($syslog_param{'priority'}, $msg);
	} else {
	    unless (flock(DATAFILE, LOCK_EX)) {
		my($msg) = "couldn't lock DB '$cache_file': $!";
	        print(STDERR "$msg\n") if ($verbose);
		syslog($syslog_param{'priority'}, $msg);
	    } else {
		if ($database{$addr}) {
		    delete($database{$addr});
		} else {
		    print(STDERR "address '$addr' not found in cache file '$cache_file'\n") if ($verbose);
		}
	    }
	    $db->sync();
	    close(DATAFILE);
	}
	undef($db);
	untie(%database);
    }
}

sub delete_cache_expired() {
    my(%cache_ttl) = @_;

    my(%cache_ttl_common);
    $cache_ttl_common{'ok'} = ($cache_ttl{'sender_ok'} > $cache_ttl{'recipient_ok'}
				? $cache_ttl{'sender_ok'} : $cache_ttl{'recipient_ok'});
    $cache_ttl_common{'fail'} = ($cache_ttl{'sender_fail'} > $cache_ttl{'recipient_fail'}
				? $cache_ttl{'sender_fail'} : $cache_ttl{'recipient_fail'});
    $cache_ttl_common{'defer'} = ($cache_ttl{'sender_defer'} > $cache_ttl{'recipient_defer'}
				? $cache_ttl{'sender_defer'} : $cache_ttl{'recipient_defer'});

    print(STDERR "delete expired records from cache file '$cache_file'\n") if ($verbose);
    my(%database);
    my($db) = tie(%database, 'DB_File', $cache_file, O_CREAT | O_RDWR, 0644);
    unless ($db) {
	my($msg) = "couldn't init DB '$cache_file': $!";
	print(STDERR "$msg\n") if ($verbose);
	syslog($syslog_param{'priority'}, $msg);
    } else {
	my($fd) = $db->fd();
	unless (open(DATAFILE, "+<&=$fd")) {
	    my($msg) = "couldn't safety open DB '$cache_file': $!";
	    print(STDERR "$msg\n") if ($verbose);
	    syslog($syslog_param{'priority'}, $msg);
	} else {
	    unless (flock(DATAFILE, LOCK_EX)) {
		my($msg) = "couldn't lock DB '$cache_file': $!";
	        print(STDERR "$msg\n") if ($verbose);
		syslog($syslog_param{'priority'}, $msg);
	    } else {
		my($deleted);
		foreach my $key (keys(%database)) {
		    my($value) = $database{$key};
		    my($year, $month, $mday, $hour, $min, $sec, $command, $answer);
		    if (($year, $month, $mday, $hour, $min, $sec, $command, $answer) =
			($value =~ /^\s*(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})\s+(.+?)\t(.+)/)) {
			$year = $year - 1900;
			$month--;
		        my($datetime) = timelocal($sec, $min, $hour, $mday, $month, $year);
			($sec, $min, $hour, $mday, $month, $year) = (localtime)[0,1,2,3,4,5];
		        my($ttl);
			if ($answer =~ /^2\d\d\s/) {
			    $ttl = $cache_ttl_common{'ok'};
			} elsif ($answer =~ /^4\d\d\s/) {
			    $ttl = $cache_ttl_common{'defer'};
			} elsif ($answer =~ /^5\d\d\s/) {
			    $ttl = $cache_ttl_common{'fail'};
			} else {
			    $ttl = 0;
			}
			if (timelocal($sec, $min, $hour, $mday, $month, $year) - $datetime > $ttl) {
			    delete($database{$key});
			    $deleted++;
			}
		    }
		}
		if ($deleted == 1) {
		    print(STDERR "$deleted expired record has been deleted from cache file '$cache_file'\n") if ($verbose);
		} elsif ($deleted > 1) {
		    print(STDERR "$deleted expired records have been deleted from cache file '$cache_file'\n") if ($verbose);
		} else {
		    print(STDERR "nothing to delete from cache file '$cache_file'\n") if ($verbose);
		}
	    }
	    $db->sync();
	    close(DATAFILE);
	}
	undef($db);
	untie(%database);
    }
}

sub dump_cache() {
    print(STDERR "dump cache file '$cache_file'\n") if ($verbose);
    my(%database);
    my($db) = tie(%database, 'DB_File', $cache_file, O_RDWR, 0444);
    unless ($db) {
	my($msg) = "couldn't init DB '$cache_file': $!";
	print(STDERR "$msg\n") if ($verbose);
	syslog($syslog_param{'priority'}, $msg);
    } else {
	my($fd) = $db->fd();
	unless (open(DATAFILE, "+<&=$fd")) {
	    my($msg) = "couldn't safety open DB '$cache_file': $!";
	    print(STDERR "$msg\n") if ($verbose);
	    syslog($syslog_param{'priority'}, $msg);
	} else {
	    unless (flock(DATAFILE, LOCK_SH)) {
		my($msg) = "couldn't lock DB '$cache_file': $!";
	        print(STDERR "$msg\n") if ($verbose);
		syslog($syslog_param{'priority'}, $msg);
	    } else {
		foreach my $key (keys(%database)) {
		    print("$key\t".$database{$key}."\n");
		}
	    }
	    $db->sync();
	    close(DATAFILE);
	}
	undef($db);
	untie(%database);
    }
}

sub get_cached_record() {
    my($address, $addr_type) = @_;
    $address =~ s/^\s*\<//;
    $address =~ s/\>\s*$//;

    my($command, $answer);
    my(%database);
    my($db) = tie(%database, 'DB_File', $cache_file, O_CREAT | O_RDWR, 0444);
    unless ($db) {
	my($msg) = "couldn't init DB '$cache_file': $!";
	print(STDERR "$msg\n") if ($verbose);
	syslog($syslog_param{'priority'}, $msg);
    } else {
	my($fd) = $db->fd();
	unless (open(DATAFILE, "+<&=$fd")) {
	    my($msg) = "couldn't safety open DB '$cache_file': $!";
	    print(STDERR "$msg\n") if ($verbose);
	    syslog($syslog_param{'priority'}, $msg);
	} else {
	    unless (flock(DATAFILE, LOCK_SH)) {
		my($msg) = "couldn't lock DB '$cache_file': $!";
	        print(STDERR "$msg\n") if ($verbose);
		syslog($syslog_param{'priority'}, $msg);
	    } else {
		my($value) = $database{$address};
		my($year, $month, $mday, $hour, $min, $sec);
		($year, $month, $mday, $hour, $min, $sec, $command, $answer) =
		    ($value =~ /^\s*(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})\s+(.+?)\t(.+)/);
		if ($answer) {
		    $year = $year - 1900;
		    $month--;
		    my($datetime) = timelocal($sec, $min, $hour, $mday, $month, $year);
		    ($sec, $min, $hour, $mday, $month, $year) = (localtime)[0,1,2,3,4,5];

		    my($ttl);
		    if ($addr_type eq 'mail') {
			if ($answer =~ /^2\d\d\s/) {
			    $ttl = $cache_ttl{'sender_ok'};
			} elsif ($answer =~ /^4\d\d\s/) {
			    $ttl = $cache_ttl{'sender_defer'};
			} elsif ($answer =~ /^5\d\d\s/) {
			    $ttl = $cache_ttl{'sender_fail'};
			} else {
			    $ttl = 0;
			}
		    } elsif ($addr_type eq 'rcpt') {
			if ($answer =~ /^2\d\d\s/) {
			    $ttl = $cache_ttl{'recipient_ok'};
			} elsif ($answer =~ /^4\d\d\s/) {
			    $ttl = $cache_ttl{'recipient_defer'};
			} elsif ($answer =~ /^5\d\d\s/) {
			    $ttl = $cache_ttl{'recipient_fail'};
			} else {
			    $ttl = 0;
			}
		    } else {
			$ttl = 0;
		    }

		    if (timelocal($sec, $min, $hour, $mday, $month, $year) - $datetime > $ttl) {
			$command = '';
			$answer = '';
		    }
		}
	    }
	    close(DATAFILE);
	}
	undef($db);
	untie(%database);
    }
    return(($command, $answer));
}

sub connect_callback {
    my $ctx = shift;
    my $hostname = shift;
    my $sockaddr_in = shift;

    my($callback) = 'connect_callback';
    my($ip_addr);

    if (($skip_relay_from) or ($skip_hosts)) {
	my($port, $iaddr) = sockaddr_in($sockaddr_in);
        $ip_addr = inet_ntoa($iaddr);
	$ip_addr = '' unless ($ip_addr);
	$ip_addr = '127.0.0.1' if (!$ip_addr and ($hostname =~ /^localhost$/i));

	print(STDERR "$callback: got connect from $hostname [$ip_addr]\n") if ($verbose);
    }

    if ($skip_relay_from) {
	my($qualifier) = 'Connect';
	my($key, $value);
	($key, $value) = &lookup_access_db($ip_addr, $qualifier, 'ip') unless ($value);
	($key, $value) = &lookup_access_db($hostname, $qualifier, 'hostname') unless ($value);
	$qualifier = '' unless ($value);
	($key, $value) = &lookup_access_db($ip_addr, '', 'ip') unless ($value);
	($key, $value) = &lookup_access_db($hostname, '', 'hostname') unless ($value);

	my($href) = $ctx->getpriv();
        $$href{'relayfrom'} = ($value =~ /^(RELAY)/i ? 1 : 0);
	$ctx->setpriv($href);

	if ($$href{'relayfrom'}) {
	    print(STDERR "$callback: relay from connection: '".
		($qualifier ? "$qualifier:" : '')."$key' => '$value'\n") if ($verbose);
	}
    }
    if ($skip_hosts) {
	my($qualifier) = 'VerifySenderHost';
	my($key, $value);
	($key, $value) = &lookup_access_db($ip_addr, $qualifier, 'ip') unless ($value);
	($key, $value) = &lookup_access_db($hostname, $qualifier, 'hostname') unless ($value);

	my($href) = $ctx->getpriv();
        $$href{'skip_hosts'} = ($value =~ /^(SKIP|OK)/i ? 1 : 0);
        $$href{'skip_hosts_record'} = "$qualifier:$key $value";
	$ctx->setpriv($href);
    }

    return SMFIS_CONTINUE;
}

sub verify_sender {
    my($mail_from, $callback, $relay_from, $auth_authen, $auth_type, $skip_hosts, $skip_hosts_record) = @_;

    print(STDERR "$callback: got mail from:$mail_from\n") if ($verbose);

    if ($mail_from =~ /^\s*\<\>\s*$/) {
	print(STDERR "$callback: skip null sender address\n") if ($verbose);
    } elsif ($skip_authen and $auth_type) {
	print(STDERR "$callback: got auth_type: '$auth_type'\n") if ($verbose);
	print(STDERR "$callback: skip verify sender address for authenticated sender '$auth_authen'\n") if ($verbose);
    } elsif ($relay_from) {
	print(STDERR "$callback: skip verify sender address for 'relay from' message\n") if ($verbose);
    } elsif ($skip_hosts) {
	print(STDERR "$callback: skip verify sender address due to the record in access_db: $skip_hosts_record\n") if ($verbose);
    } else {
	$mail_from =~ s/\s*\<(.+?)\>\s*/$1/;
	my($qualifier) = 'VerifySender';
	my($key, $db_value) = &lookup_access_db($mail_from, $qualifier, 'mail');
	if ($db_value =~ /^(OK|SKIP)/i) {
	    print(STDERR "$callback: skip verify sender address due to the record in access_db: '$qualifier:$key' => '$db_value'\n") if ($verbose);
        } else {
	    print(STDERR "$callback: find address '$mail_from' in cache\n") if ($verbose);
	    my($command, $answer) = &get_cached_record($mail_from, 'mail');
	    if ($answer) {
    	        print(STDERR "$callback: found in cache command: $command\n") if ($verbose);
		print(STDERR "$callback: found in cache answer:  $answer\n") if ($verbose);
	    } else {
		print(STDERR "$callback: status of address '$mail_from' not found in cache\n") if ($verbose);
    		($command, $answer) = verify_address($mail_from, $use_vrfy, $max_attempts, \%syslog_param, $verbose);

		$command =~ s/[\r\n]+$//;
	        $answer =~ s/[\r\n]+$//;
    		$mail_from =~ s/^\s*\<//;
		$mail_from =~ s/\>\s*$//;

		&save_cache($mail_from, "$command\t$answer");

    	        print(STDERR "$callback: returned command: $command\n") if ($verbose);
		print(STDERR "$callback: returned answer:  $answer\n") if ($verbose);
	    }

    	    my($error_code, $enh_error_code, $error_msg, $reply);
	    if ($answer =~ /^(\d\d\d)\s+(\d\.\d\.\d)\s+(.+)$/) {
	        $error_code = $1;
	        $enh_error_code = $2;
	        $error_msg = $3;
	    } elsif ($answer =~ /^(\d\d\d)\s+(.+)$/) {
	        $error_code = $1;
	        $error_msg = $2;
	        $enh_error_code = '';
	    }
    	    $error_msg =~ s/[\r\n]$//;
	    if ($error_code =~ /^5\d\d/) {
	        $reply = "Sender verify fail for <$mail_from>: response to \"$command\" was $answer";
	        print(STDERR "$callback: setreply: '550', '5.7.1', '$reply'\n") if ($verbose);
	        return((SMFIS_REJECT, '550', '5.7.1', $reply));
	    } elsif ($error_code =~ /^4\d\d/) {
		if ($db_value =~ /^DEFER_OK$/i) {
		    print(STDERR "$callback: skip verify sender address due to the record in access_db: '$qualifier:$key' => '$db_value'\n") if ($verbose);
		} elsif ($sender_tempfail_accept) {
		    print(STDERR "$callback: skip verify sender address due to the local policy\n") if ($verbose);
		} else {
	    	    $reply = "Could not complete sender verify callout for <$mail_from>: $answer";
	    	    print(STDERR "$callback: setreply: '451', '4.4.0', '$reply'\n") if ($verbose);
	    	    return((SMFIS_TEMPFAIL, '451', '4.4.0', $reply));
		}
	    } elsif ($error_code !~ /^2\d\d/) {
	        $reply = "Sender verify fail for <$mail_from>: $answer";
	        print(STDERR "$callback: setreply: '550', '5.7.1', '$reply'\n") if ($verbose);
	        return((SMFIS_REJECT, '550', '5.7.1', $reply));
	    }
	}
    }
    return SMFIS_CONTINUE;
}

sub mailfrom_callback {
    my $ctx = shift;
    my $mail_from = shift;
    $mail_from =~ s/<.+(\<.+?\>)\>/$1/;

    my($href) = $ctx->getpriv();
    $$href{'auth_authen'} = $ctx->getsymval('{auth_authen}');
    $$href{'auth_type'} = $ctx->getsymval('{auth_type}');
    $$href{'mail_from'} = $mail_from;
    $ctx->setpriv($href);

    if ($delay_checks) {
	print(STDERR "mailfrom_callback: verify sender for '$mail_from' delayed\n") if ($verbose);
    } else {
	my($result, $error_code, $error_code_enh, $reply) =
	    &verify_sender($mail_from, 'mailfrom_callback',
	    $$href{'relayfrom'}, $$href{'auth_authen'}, $$href{'auth_type'}, $$href{'skip_hosts'}, $$href{'skip_hosts_record'});
	if ($result != SMFIS_CONTINUE) {
    	    $ctx->setreply($error_code, $error_code_enh, $reply);
	    return $result;
	}
    }
    return SMFIS_CONTINUE;
}

sub rcptto_callback {
    my $ctx = shift;
    my $rcpt_to = shift;
    $rcpt_to =~ s/\s*\<(.+?)\>\s*/$1/;

    my($href) = $ctx->getpriv();
    $ctx->setpriv($href);

    print(STDERR "rcptto_callback: got rcpt to:$rcpt_to\n") if ($verbose);

    if ($skip_authen and $$href{'auth_type'}) {
	print(STDERR "rcptto_callback: got auth_type: '".$$href{'auth_type'}."'\n") if ($verbose);
	print(STDERR "rcptto_callback: skip verify recipient address for authenticated sender '".
	    $$href{'auth_authen'}."'\n") if ($verbose);
    } elsif ($$href{'relayfrom'}) {
	print(STDERR "rcptto_callback: skip verify recipient address for 'relay from' message\n") if ($verbose);
    } else {
	my($qualifier) = 'VerifyRecipient';
	my($key, $db_value) = &lookup_access_db($rcpt_to, $qualifier, 'mail');
	if ($db_value =~ /^(OK|SKIP)/i) {
	    print(STDERR "rcptto_callback: skip verify recipient address due to the record in access_db: '$qualifier:$key' => '$db_value'\n") if ($verbose);
        } else {
	    print(STDERR "rcptto_callback: find address '$rcpt_to' in cache\n") if ($verbose);
	    my($command, $answer) = &get_cached_record($rcpt_to, 'rcpt');
	    if ($answer) {
    	        print(STDERR "rcptto_callback: found in cache command: $command\n") if ($verbose);
		print(STDERR "rcptto_callback: found in cache answer:  $answer\n") if ($verbose);
	    } else {
		print(STDERR "rcptto_callback: status of address '$rcpt_to' not found in cache\n") if ($verbose);
		if ($db_value =~ /^\s*exchange\s*:\s*(.+)$/i) {
		    my(@param) = split(/\s*:\s*/, $1);
		    use check_mail_address_exchange;
    		    ($command, $answer) = verify_address_exchange($rcpt_to, $param[0], $param[1], $param[2], $param[3], \%syslog_param, $verbose);
		    $db_value = 'DEFER_OK' if ($db_value =~ /defer_ok/i);
		} else {
    		    ($command, $answer) = verify_address($rcpt_to, $use_vrfy, $max_attempts, \%syslog_param, $verbose);
		}

		$command =~ s/[\r\n]+$//;
	    	$answer =~ s/[\r\n]+$//;
    		$rcpt_to =~ s/^\s*\<//;
	    	$rcpt_to =~ s/\>\s*$//;

    		&save_cache($rcpt_to, "$command\t$answer");

    	        print(STDERR "rcptto_callback: returned command: $command\n") if ($verbose);
		print(STDERR "rcptto_callback: returned answer:  $answer\n") if ($verbose);
	    }

    	    my($error_code, $enh_error_code, $error_msg, $reply);
	    if ($answer =~ /^(\d\d\d)\s+(\d\.\d\.\d)\s+(.+)$/) {
		$error_code = $1;
	        $enh_error_code = $2;
		$error_msg = $3;
	    } elsif ($answer =~ /^(\d\d\d)\s+(.+)$/) {
		$error_code = $1;
		$error_msg = $2;
		$enh_error_code = '';
	    }
    	    $error_msg =~ s/[\r\n]$//;
	    if ($error_code =~ /^5\d\d/) {
	    	$error_code = '554';
		$enh_error_code = '5.1.1';
	        $reply = 'Unroutable address';
	        print(STDERR "rcptto_callback: setreply: '$error_code', '$enh_error_code', '$reply'\n") if ($verbose);
    	        $ctx->setreply($error_code, $enh_error_code, $reply);
	        return SMFIS_REJECT;
	    } elsif ($error_code =~ /^4\d\d/) {
		if ($db_value =~ /^DEFER_OK$/i) {
		    print(STDERR "rcptto_callback: skip verify recipient address due to the record in access_db: '$qualifier:$key' => '$db_value'\n") if ($verbose);
		} elsif ($rcpt_tempfail_accept) {
		    print(STDERR "rcptto_callback: skip verify recipient address due to the local policy\n") if ($verbose);
		} else {
	    	    $reply = "Could not complete recipient verify callout for <$rcpt_to>: $answer";
		    print(STDERR "rcptto_callback: setreply: '451', '4.4.0', '$reply'\n") if ($verbose);
    	    	    $ctx->setreply('451', '4.5.1', $reply);
    		    return SMFIS_TEMPFAIL;
		}
    	    } elsif ($error_code !~ /^2\d\d/) {
	    	$error_code = '554';
		$enh_error_code = '5.1.1';
	    	$reply = 'Unroutable address';

		print(STDERR "rcptto_callback: setreply: '$error_code', '$enh_error_code', '$reply'\n") if ($verbose);
    	        $ctx->setreply($error_code, $enh_error_code, $reply);
		return SMFIS_REJECT;
	    }
	}
    }
    if ($delay_checks) {
	my($result, $error_code, $error_code_enh, $reply) =
	    &verify_sender($$href{'mail_from'}, 'rcptto_callback (delayed verify sender)',
	    $$href{'relayfrom'}, $$href{'auth_authen'}, $$href{'auth_type'}, $$href{'skip_hosts'}, $$href{'skip_hosts_record'});
	if ($result != SMFIS_CONTINUE) {
    	    $ctx->setreply($error_code, $error_code_enh, $reply);
	    return $result;
	}
    }

    return SMFIS_CONTINUE;
}

sub abort_callback {
    my $ctx = shift;
    $ctx->setpriv(undef);
    return SMFIS_CONTINUE;
}

sub close_callback {
    my $ctx = shift;
    $ctx->setpriv(undef);
    return SMFIS_CONTINUE;
}

my %my_callbacks = (
    'connect' =>	\&connect_callback,
#    'helo'	=> \&helo_callback,
    'envfrom'	=> \&mailfrom_callback,
    'envrcpt'	=> \&rcptto_callback,
#    'header'	=> \&header_callback,
#    'eoh'	=> \&eoh_callback,
#    'body'	=> \&body_callback,
#    'eom'	=> \&eom_callback,
    'abort'	=> \&abort_callback,
    'close'	=> \&close_callback,
);

BEGIN: {
    my($milter_name, $sendmail_config, $help, $version);

    #    
    GetOptions (
        'help'				=> \$help,
        'version'			=> \$version,
        'milter-name=s'			=> \$milter_name,
        'sendmail-config=s'		=> \$sendmail_config,
        'syslog-facility=s'		=> \$syslog_param{'facility'},
        'verbose'			=> \$verbose,
	'use-vrfy'			=> \$use_vrfy,
        'max-attempts=i'		=> \$max_attempts,
        'rcpt-tempfail=s'		=> \$rcpt_tempfail_accept,
        'sender-tempfail=s'		=> \$sender_tempfail_accept,
        'access_db=s'			=> \$access_db,
	'skip-authen'			=> \$skip_authen,
	'skip-relay-from'		=> \$skip_relay_from,
	'skip-hosts'			=> \$skip_hosts,
	'cache-file=s'			=> \$cache_file,
	'cache-ttl-sender-ok=i'		=> \$cache_ttl{'sender_ok'},
	'cache-ttl-sender-fail=i'	=> \$cache_ttl{'sender_fail'},
	'cache-ttl-sender-defer=i'	=> \$cache_ttl{'sender_defer'},
	'cache-ttl-recipient-ok=i'	=> \$cache_ttl{'recipient_ok'},
	'cache-ttl-recipient-fail=i'	=> \$cache_ttl{'recipient_fail'},
	'cache-ttl-recipient-defer=i'	=> \$cache_ttl{'recipient_defer'},
	'clear-cache'           	=> \$clear_cache,
	'delete-cache-record=s' 	=> \$delete_cache,
	'delete-cache-expired'  	=> \$delete_cache_expired,
	'dump-cache'			=> \$dump_cache,
	'delay-checks'			=> \$delay_checks
    );

    if ($help) {
        print(
"$info

Usage: perl $script_name [OPTIONS]

General options:
  --milter-name                     name of mail filter in sendmail.cf
  --sendmail-config                 path to the sendmail.cf

Cache options:
  --dump-cache                      dump cache file
  --clear-cache                     clear cache
  --delete-cache-record=address     delete address from cache
  --delete-cache-expired            delete expired records from cache

Additional options:
  --syslog-facility                 syslog facility (default local4)
  --verbose                         be verbose
  --use-verify                      try to use VRFY SMTP command to check
                                    address
  --max-attempts=<count>            how long try to connect to best MX for
                                    verifing adrress (default 1)
  --sender-tempfail=accept|tempfail default behaviour for verify sender temporary
                                    failed (default ".($defaults{'sender_tempfail_accept'} ? "accept" : "tempfail").")
  --rcpt-tempfail=accept|tempfail   default behaviour for  verify recipient temporary
                                    failed (default ".($defaults{'rcpt_tempfail_accept'} ? "accept" : "tempfail").")
  --access_db=file1[:file2[:file3]] access_db files for checking excludes
                                    (default /etc/mail/access.db)
  --skip-authen                     skip checking for authenticated senders
  --skip-relay-from                 skip checking for outgoing messages
  --skip-hosts                      skip verify sender from hosts which found
                                    in access_db with qualifier VerifySenderHost
				    and value SKIP or OK
  --delay-checks                    verify sender in RCPT TO callback

Additional cache options:
  --cache-file                      default ".$defaults{'cache_file'}."
  --cache-ttl-sender-ok=ttl         time to live (in sec) for successful verify
                                    sender cached records (default ".$defaults_ttl{'sender_ok'}.")
  --cache-ttl-sender-fail=ttl       time to live (in sec) for failed verify
                                    sender cached records (default ".$defaults_ttl{'sender_fail'}.")
  --cache-ttl-sender-defer=ttl      time to live (in sec) for defered verify
                                    sender cached records (default ".$defaults_ttl{'sender_defer'}.")
  --cache-ttl-recipient-ok=ttl      time to live (in sec) for successful verify
                                    recipient cached records (default ".$defaults_ttl{'recipient_ok'}.")
  --cache-ttl-recipient-fail=ttl    time to live (in sec) for failed verify
                                    sender recipient records (default ".$defaults_ttl{'recipient_fail'}.")
  --cache-ttl-recipient-defer=ttl   time to live (in sec) for defered verify
                                    sender recipient records (default ".$defaults_ttl{'recipient_defer'}.")

Miscellaneous:
  --version                         display the version of $script_name and
                                    exit
  --help                            display this help and exit
");
        exit();
    }
    if ($version) {
        print("$info\n");
        exit();
    }

    $cache_file = $defaults{'cache_file'} unless ($cache_file);
    foreach (keys(%defaults_ttl)) {
	$cache_ttl{$_} = $defaults_ttl{$_} unless ($cache_ttl{$_});
    }

    if ($clear_cache) {
	&clear_cache();
        exit();
    }

    if ($delete_cache) {
	&delete_cache($delete_cache);
        exit();
    }

    if ($delete_cache_expired) {
	&delete_cache_expired(%cache_ttl);
        exit();
    }

    if ($dump_cache) {
	&dump_cache();
        exit();
    }

    $syslog_param{'facility'} = $defaults{'syslog_facility'} unless ($syslog_param{'facility'});
    $max_attempts = $defaults{'max_attempts'} unless ($max_attempts);
    $use_vrfy = $defaults{'use_vrfy'} unless ($use_vrfy eq '');
    $access_db = $defaults{'access_db'} unless ($access_db);

    if ($sender_tempfail_accept =~ /^accept$/) {
	$sender_tempfail_accept = 1;
    } elsif ($sender_tempfail_accept =~ /^tempfail$/) {
	$sender_tempfail_accept = 0;
    } else {
	$sender_tempfail_accept = $defaults{'sender_tempfail_accept'};
    }

    if ($rcpt_tempfail_accept =~ /^accept$/) {
	$rcpt_tempfail_accept = 1;
    } elsif ($rcpt_tempfail_accept =~ /^tempfail$/) {
	$rcpt_tempfail_accept = 0;
    } else {
	$rcpt_tempfail_accept = $defaults{'rcpt_tempfail_accept'};
    }

    #    sendmail',        
    $sendmail_config = find_sendmail_config(split(':', $defaults{'sendmail_config'})) unless ($sendmail_config);

    #     ,       
    unless ($milter_name and $sendmail_config) {
	print("name of mail filter in sendmail.cf is missing") unless ($milter_name);
        print("path to the sendmail.cf is missing") unless ($sendmail_config);
        print("

Usage: perl $script_name [OPTIONS]

Try `$script_name --help' for more options.

");
        exit();
    }

    setlogsock($syslog_param{'socket'});
    openlog($syslog_param{'ident'}, $syslog_param{'options'}, $syslog_param{'facility'});

    my($conn) = Sendmail::Milter::auto_getconn($milter_name, $sendmail_config);

    print_msg("Found connection info for '$milter_name': $conn");

    if ($conn =~ /^local:(.+)$/) {
	my $unix_socket = $1;
	if (-e $unix_socket) {
	    print_msg("Attempting to unlink UNIX socket '$conn'...");
	    if (unlink($unix_socket) == 0) {
		print_msg("failed");
		exit();
	    }
	    print_msg("successful");
	}
    }

    if (not Sendmail::Milter::auto_setconn($milter_name, $sendmail_config)) {
	print_msg("Failed to detect connection information");
    } elsif (not Sendmail::Milter::register($milter_name, \%my_callbacks, SMFI_CURR_ACTS)) {
	print_msg("Failed to register callbacks for '$milter_name'");
    } else {
	print_msg("Starting Sendmail::Milter $Sendmail::Milter::VERSION engine");
	if (Sendmail::Milter::main()) {
	    print_msg("Successful exit from the Sendmail::Milter engine");
	} else {
	    print_msg("Unsuccessful exit from the Sendmail::Milter engine");
	}
    }
    closelog();
}

sub print_msg() {
    my($msg) = @_;
    print($msg."\n");
    syslog('notice', $msg);
}

sub find_sendmail_config {
    my(@sendmail_config_defaults) = @_;
    my($file);
    foreach $file (@sendmail_config_defaults) {
        if (-f $file) {
            return($file);
        }
    }
    return('');
}
